// Copyright 2017 - 2018 Modoolar <info@modoolar.com>
// License LGPLv3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html).

odoo.define('scrummer.view.backlog', function (require) {
    "use strict";
    const data = require('scrummer.data');
    const DataServiceFactory = require('scrummer.data_service_factory');
    const AgileViewWidget = require('scrummer.BaseWidgets').AgileViewWidget;
    const AgileModals = require('scrummer.widget.modal');
    const AbstractModelList = require('scrummer.abstract_model_list');
    const ModelList = require('scrummer.model_list');
    const TaskWidget = require('scrummer.widget.task').TaskWidget;
    const hash_service = require('scrummer.hash_service');
    const pluralize = require('pluralize');
    const web_core = require('web.core');
    const qweb = web_core.qweb;
    const _t = web_core._t;
    const core = require('scrummer.core');
    const AgileToast = require('scrummer.toast');
    const Sortable = require('sortable');
    const dialog = require('scrummer.dialog');
    const mixins = require('scrummer.mixins');
    const Waypoint = window.Waypoint;
    const TOP = _t("Top");
    const BOTTOM = _t("Bottom");

    const task_service = DataServiceFactory.get("project.task", false);
    const project_service = DataServiceFactory.get("project.project");

    const SimpleBacklogTaskItem = ModelList.SimpleTaskItem.extend(mixins.MenuItemsMixin, {
        //order_field: "agile_order",
        _name: "BacklogTaskItem",
        events: {
            'click': function (evt) {
                if (evt.isDefaultPrevented()) {
                    // Skip if default behaviour is prevented, eg. when clicked on menu.
                    return;
                }
                this._onItemClicked(evt);
            },
        },
        menuItems: [
            {
                class: "assign-to-me",
                icon: "mdi-account-check",
                image: function () {
                    return this.currentUser.imageUrl;
                },
                text: _t("Assign To Me"),
                callback: '_onAssignToMeClick',
                sequence: 1,
                hidden() {
                    return this.record.user_id && this.record.user_id[0] === data.session.uid;
                }
            },
            {
                class: "unassign",
                icon: "mdi-account-minus",
                text: _t("Unassign"),
                callback: '_onUnassignClick',
                sequence: 2,
                hidden() {
                    return !(this.record.user_id && this.record.user_id[0] === data.session.uid);
                }
            },
            {
                class: "edit-item",
                icon: "mdi-pencil",
                text: _t("Edit"),
                callback: '_onEditItemClick',
                sequence: 3,
            },
            {
                class: "add-sub-item",
                icon: "mdi-subdirectory-arrow-right",
                text: _t("Add Sub Item"),
                callback: '_onAddSubItemClick',
                sequence: 4,
                hidden() {
                    return !this.task_type.allow_sub_tasks;
                },
            },
            {
                class: "add-link",
                icon: "mdi-link",
                text: _t("Add Link"),
                callback: '_onAddLinkClick',
                sequence: 5,
            },
            {
                class: "add-comment",
                icon: "mdi-comment-account",
                text: _t("Add Comment"),
                callback: '_onAddCommentClick',
                sequence: 6,
            },
            {
                class: "delete",
                icon: "mdi-delete",
                text: _t("Delete"),
                callback: '_onDelete',
                sequence: 7,
            },
        ],
        init() {
            this._super.apply(this, arguments);
            mixins.MenuItemsMixin.init.call(this);
        },
        start() {
            mixins.MenuItemsMixin.start.call(this);
            this.$(".task-menu, .dropdown-content a").click((evt) => {
                // Prevent triggering click handler for entire task item
                evt.preventDefault();
            });
            return this._super();
        },
        addedToDOM() {
            this._super();
            this.$('.dropdown-button').dropdown();

        },
        rerenderWidget() {
            this.renderElement();
            this.start();
            this.addedToDOM();
        },
        _onItemClicked() {
            this.trigger_up("open_right_side", {
                WidgetClass: TaskWidget,
                options: {id: this.record.id, isQuickDetailView: true}
            });
        },
        _onAssignToMeClick() {
            this.record.user_id = data.session.uid;
        },
        _onUnassignClick() {
            this.record.user_id = false;
        },
        _onEditItemClick() {
            this.trigger_up("open_new_item_modal", {
                currentProjectId: this.record.project_id[0],
                focus: "name",
                edit: this.record,
            });
        },
        _onAddSubItemClick() {
            this.trigger_up("open_new_item_modal", {
                parent_id: this.record.id,
                currentProjectId: this.record.project_id[0]
            });
        },
        _onAddLinkClick() {
            this.trigger_up("open_link_modal", {id: this.record.id});
        },
        _onAddCommentClick() {
            const modal = new AgileModals.CommentItemModal(this, {
                task: this.record,
            });
            modal.appendTo($("body"));
        },
        _onDelete() {
            dialog.confirm(_t("Delete task"), _t("Are you sure you want to delete this task?"), _t("yes")).done(() => {
                this.record.unlink();
            });
        },
    });

    const BacklogList = AbstractModelList.ModelList.extend({
        _name: "BacklogList",
        pageSize: 10,
        init(parent, options) {
            this._super(parent, options);
            this.size = this.pageSize;
            this._require_prop("task_ids", "Field task_ids is array containing ids of all tasks in backlog");
            this._require_prop("taskWidgetItemCache", "JavaScript Map object where key is id, and value is cached widget item.");
            this._require_prop("_getNewListWidget", "A function that accepts list id in and returns ModelList widget instance");
            this.allFilteredTaskIds = this.getAllTaskIds();
        },
        getAllTaskIds() {
            return this.task_ids;
        },
        loadItems() {
            return task_service.getRecords(this.getSlicedBacklogTaskIds(this.getFilteredTaskIds(this.allFilteredTaskIds)))
                .then((tasks) => {
                    this.data = tasks;
                });
        },
        shouldLoadMore() {
            return this.allFilteredTaskIds.length > this.list.size;
        },
        getFilteredTaskIds(filtered_task_ids) {
            return filtered_task_ids
                ? this.getAllTaskIds().filter((id) => filtered_task_ids.includes(id))
                : this.getAllTaskIds();
        },
        getSlicedBacklogTaskIds(backlog_task_ids) {
            return backlog_task_ids.slice(0, this.size);
        },
        loadMoreItems() {
            this.size += 10;
            const sliced_task_ids = this.getSlicedBacklogTaskIds(this.getFilteredTaskIds(this.allFilteredTaskIds));
            this.trigger_up("loading_more_items_started");
            task_service.getRecords(sliced_task_ids).then((tasks) => {
                this.data = tasks;
                return tasks;
            }).then((tasks) => {
                tasks.forEach(this.addItem.bind(this));
                this.trigger_up("loading_more_items_finished");

                /* Because lazy loading backlog items work when you scroll down to bottom of backlog,
                * we need to manage the case where bottom of backlog is above bottom of window.
                * In such case, when master list is less then 110% in height, load more items if available
                */
                this.trigger_up("backlog_list_changed", {
                    size: this.list.size,
                    total: this.allFilteredTaskIds.length
                });
            });
        },
        addItem(item) {
            if (this.list.has(item.id)) {
                return;
            }
            !this.getAllTaskIds().includes(item.id) && this.getAllTaskIds().push(item.id);
            !this.allFilteredTaskIds.includes(item.id) && this.allFilteredTaskIds.push(item.id);

            const cachedWidget = this.taskWidgetItemCache.get(item.id);
            // Use cached widget if exists on backlog view, or create it and store in cache
            const retVal = this._super(cachedWidget || item);
            cachedWidget || this.taskWidgetItemCache.set(item.id, retVal);

            this.trigger_up("backlog_list_changed", {
                size: this.list.size,
                total: this.allFilteredTaskIds.length
            });
            return retVal;
        },
        removeItem(id) {
            const removed = this._super.apply(this, arguments);
            this.allFilteredTaskIds.includes(id) && this.allFilteredTaskIds.splice(this.allFilteredTaskIds.indexOf(id), 1);
            if (removed) {
                this.trigger_up("backlog_list_changed", {
                    size: this.list.size,
                    total: this.allFilteredTaskIds.length
                });
            }
            return removed;
        },
        addedToDOM() {
            this.$el.waypoint({
                handler: (direction) => {
                    if (direction === "down" && this.shouldLoadMore()) {
                        // console.log("Backlog bottom hit, loading more tasks");
                        this.loadMoreItems();
                    }
                },
                context: "#backlog-view",
                offset: 'bottom-in-view'
            });
        },
        applyFilter(filteredTaskIds) {
            // Reset backlog length and slice tasks so that only first page gets loaded.
            this.size = 10;
            this.allFilteredTaskIds = this.getFilteredTaskIds(filteredTaskIds);
            // Remove filtered tasks from list
            for (const id of this.list.keys()) {
                if (!this.allFilteredTaskIds.includes(id)) {
                    this.removeItem(id, false);
                }
            }
            if (this.shouldLoadMore() && !this.loadingMoreItems) {
                this.loadMoreItems();
            }
        }
    });

    const NonBacklogList = AbstractModelList.ModelList.extend({
        _name: "NonBacklogList",
        init() {
            this._super.apply(this, arguments);
            this._require_prop("_getNewListWidget", "A function that accepts list id in and returns ModelList widget instance");
            this._require_prop("taskWidgetItemCache", "JavaScript Map object where key is id, and value is cached widget item.");
            this._require_prop("allFilteredTaskIds", "This is the array of ids that will be used for general filtering of tasks in list.");
        },
        loadItems() {
            return task_service.getRecords(this.getAllTaskIds()).then((tasks) => {
                this.data = tasks;
            });
        },
        getAllTaskIds() {
            if (!this.generalFilteredTaskIds) {
                this.generalFilteredTaskIds = this.task_ids.filter((id) => this.allFilteredTaskIds.includes(id));
            }
            return this.generalFilteredTaskIds;
        },
        getFilteredTaskIds(filtered_task_ids) {
            const allTaskIds = this.getAllTaskIds();
            return typeof filtered_task_ids === "undefined"
                ? allTaskIds
                : filtered_task_ids.filter((id) => allTaskIds.includes(id));
        },
        shouldTaskBeAdded(item) {
            if (hash_service.get("project")) {
                const project_id = parseInt(hash_service.get("project"), 10);
                const task_project_id = item._class === "ModelListItem" ? item.record.project_id : item.project_id;
                return isNaN(project_id) || (task_project_id && project_id === task_project_id[0]);
            }
            return true;
        },
        addItem(item) {
            // Prevent adding tasks to list that doesn't belong to it.
            if (!this.shouldTaskBeAdded(item)) {
                return;
            }
            if (this.list.has(item.id)) {
                return this.list.get(item.id);
            }
            const cachedWidget = this.taskWidgetItemCache.get(item.id);

            // Use cached widget if exists on backlog view, or create it and store in cache
            const retVal = this._super(cachedWidget || item);
            cachedWidget || this.taskWidgetItemCache.set(item.id, retVal);
            this.trigger_up("non_backlog_list_changed", {
                size: this.list.size,
                total: (this.allFilteredTaskIds || this.getAllTaskIds()).length,
                id: this.attributes['data-id']
            });
            return retVal;
        },
        removeItem() {
            const removed = this._super.apply(this, arguments);
            if (removed) {
                this.trigger_up("non_backlog_list_changed", {
                    size: this.list.size,
                    total: (this.allFilteredTaskIds || this.getAllTaskIds()).length,
                    id: this.attributes['data-id']
                });
            }
        },
        applyFilter(filteredTaskIds) {
            this.allFilteredTaskIds = this.getFilteredTaskIds(filteredTaskIds);
            // Add tasks that passed filter to list, if they are missing
            const taskIDsToBeAdded = this.getAllTaskIds().filter((id) => this.allFilteredTaskIds.includes(id) && !this.list.has(id));
            task_service.getRecords(taskIDsToBeAdded).then((records) => records.forEach((record) => this.addItem(record)));

            // Remove filtered tasks from list
            for (const id of this.list.keys()) {
                !this.allFilteredTaskIds.includes(id) && this.removeItem(id, false);
            }
        }
    });

    const AbstractBacklogView = AgileViewWidget.extend({
        /* eslint-disable no-unused-vars */
        title: _t("Backlog"),
        template: "scrummer.view.backlog",
        _name: "BacklogView",
        menuItems: [
            {
                class: "add-new-item",
                icon: "mdi-library-books",
                text: _t("Add new item"),
                callback: '_onAddNewItemClick'
            },
        ],
        shouldRenderDragShortcuts: true,
        custom_events: Object.assign({}, AgileViewWidget.prototype.custom_events || {}, {
            drag_start: '_onDragStart',
            drag_end: '_onDragEnd',
            backlog_list_changed: '_onBacklogListChanged',
            open_link_modal: '_onOpenLinkModal',
            open_new_item_modal: '_onOpenNewItemModal',
            loading_more_items_started: '_onLoadingMoreItemsStarted',
            loading_more_items_finished: '_onLoadingMoreItemsFinished',
        }),
        init(parent, options) {
            this._super(parent, options);

            // Getting board_id from hash and fetch all project_ids from that board in order to create filter for fetching projects
            this.board_id = parseInt(hash_service.get("board"), 10);
            this.task_id = hash_service.get("task") && parseInt(hash_service.get("task"), 10);

            this.filterQuery = "";
            this.taskWidgetItemMap = new Map();
            this.backlogLength = 10;

            window.data = data;
            window.blv = this;
        },
        removeNavSearch() {
            core.bus.trigger("search:remove");
        },
        prepareUser() {
            return data.cache.get("current_user").then((user) => {
                this.user = user;
            });
        },
        prepareProjects() {
            return project_service.dataset.id_search("", this.setupProjectIds()
                .then(this.getProjectDomain.bind(this)))
                .then((ids) => project_service.getRecords(ids))
                .then((projects) => {
                    this.projects = projects;
                });
        },
        prepareBoard() {
            return DataServiceFactory.get("project.agile.board").getRecord(this.board_id).then((board) => {
                this.board = board;
            });
        },
        // All variables that will be used in getApplyFilterDomain method must be set prior to calling this._super() when
        loadDependencies() {
            return this.prepareBoard()
                .then(() => this.prepareUser())
                .then(() => this.prepareProjects());
        },
        willStart() {
            const prepared = $.when(this._super(), this.loadDependencies());

            return $.when(
                prepared.then(() => this.getFilteredTaskIds()),
                prepared.then(() => task_service.dataset.id_search(this.filterQuery, this.getBacklogTaskDomain(), false, false, "agile_order")
                    .then((ids) => {
                        this.backlogTaskIds = ids;
                    })
                ));
        },
        getBacklogTaskDomain() {
            const domain = [
                ["project_id", "in", this.project_ids],
                ["wkf_state_type", "!=", "done"]
            ];

            if (this.boardTaskTypeFilterExists()) {
                domain.push([
                    "type_id", "in", this.board.backlog_task_type_ids
                ]);
            }
            return $.when(domain);
        },
        getProjectDomain() {
            return [
                ["board_ids", "in", this.board_id],
                ["id", "in", this.project_ids],
                ["workflow_id", "!=", false]
            ];
        },
        getApplyFilterDomain() {

            const domain = [
                "|",
                "|",
                ["key", "ilike", this.filterQuery],
                ["description", "ilike", this.filterQuery],
                ["name", "ilike", this.filterQuery],
                ["project_id", "in", this.project_ids],
            ];

            if (this.boardTaskTypeFilterExists()) {
                domain.push([
                    "type_id", "in", this.board.backlog_task_type_ids
                ]);
            }

            return $.when(domain);
        },
        _getTaskItemClass() {
            throw new Error("You should return class of non-abstract BacklogTaskItem");
        },
        renderElement() {
            this._super();
            this.backlogTaskList = new BacklogList(this, {
                model: "project.task",
                taskWidgetItemCache: this.taskWidgetItemMap,
                task_ids: this.backlogTaskIds,
                _getNewListWidget: this._getNewListWidget.bind(this),
                ModelItem: this._getTaskItemClass(),
                name: "backlog",
                sortable: {group: "backlog"},
            });

            const backlogData = {
                count: "0 issues",
                estimates: {
                    todo: 0,
                    inProgress: 0,
                    done: 0
                }
            };
            this.backlogNode = $(qweb.render("scrummer.backlog", backlogData).trim());
            this.backlogNode.insertAfter(this.$("#backlog-view .section"));
            this.backlogTaskList.insertBefore(this.$(".list-preloader"));
            this.$(".list-preloader").hide();
        },
        _getNewListWidget(list_id) {
            throw new Error("This method must be implemented");
        },
        start() {
            this._is_added_to_DOM.then(() => {
                core.bus.trigger("search:show", (input) => {
                    this.applyFilter(input.val());
                });
            });
            this.bindEventListeners();
            return this._super();
        },
        bindEventListeners() {
            this.$('.tooltipped').tooltip({delay: 50});
            core.bus.on("project.task:write", this, this._onProjectTaskWrite);
            core.bus.on("project.task:create", this, this._onProjectTaskCreate);
            core.bus.on("project.task:unlink", this, this._onProjectTaskUnlink);
        },
        setNewTaskOrder(task) {
            throw new Error("Not implemented");
        },
        getFilteredTaskIds() {
            return task_service.dataset
                .id_search("", this.getApplyFilterDomain(), false, false, "agile_order").then((task_ids) => {
                    this.allFilteredTaskIds = task_ids;
                    return task_ids;
                });
        },
        applyFilter(q) {
            if (q === this.queryFilter) {
                return;
            }
            if (typeof q === "string") {
                this.filterQuery = q;
            }
            this.getFilteredTaskIds().then(this.handleFilter.bind(this));
        },
        handleFilter(task_ids) {
            this.backlogTaskList.applyFilter(task_ids);
        },
        allTaskIds() {
            return this.backlogTaskList.task_ids.slice();
        },
        getTaskListOfTaskWidget(taskWidget) {
            throw new Error("Not implemented!");
        },
        isTaskInBacklog(task) {
            throw new Error("Not implemented!");
        },
        addTaskToNonBacklogList(task) {
            throw new Error("Not implemented!");
        },
        removeTask(id, removeFromCache = false, syncerMeta) {
            const taskWidget = this.taskWidgetItemMap.get(id);
            if (!taskWidget) {
                return false;
            }
            this.getTaskListOfTaskWidget(taskWidget).removeItem(id);
            if (removeFromCache) {
                this.taskWidgetItemMap.delete(id);
            }
            if (syncerMeta) {
                if (syncerMeta.user_id.id !== data.session.uid && syncerMeta.indirect === false) {
                    AgileToast.toastTask(syncerMeta.user_id, syncerMeta.data, syncerMeta.method);
                }
            }
            return true;
        },

        boardTaskTypeFilterExists() {
            return this.board.backlog_task_type_ids && this.board.backlog_task_type_ids.length > 0;
        },

        isBoardAllowedTaskType(type_id) {
            return this.boardTaskTypeFilterExists() ? this.board.backlog_task_type_ids.includes(type_id) : true;
        },

        addTask(id, syncerMeta) {
            $.when(task_service.getRecord(id)).then((task) => {
                // Skip adding tasks from other projects
                if (!this.project_ids.includes(task.project_id[0])) {
                    return;
                }

                if (!this.isBoardAllowedTaskType(task.type_id[0])) {
                    return;
                }

                const taskWidget = this.isTaskInBacklog(task)
                    ? this.backlogTaskList.addItem(task)
                    : this.addTaskToNonBacklogList(task);

                if (!taskWidget) {
                    return;
                }
                this.taskWidgetItemMap.has(task.id) || this.taskWidgetItemMap.set(task.id, taskWidget);
                if (syncerMeta) {
                    if (syncerMeta.user_id.id !== data.session.uid && syncerMeta.indirect === false) {
                        AgileToast.toastTask(syncerMeta.user_id, task, syncerMeta.method);
                    }
                }
            });
        },
        setupProjectIds() {
            return data.cache.get("current_user").then((user) => {
                const hashProjectId = parseInt(hash_service.get("project"), 10);
                if (typeof hash_service.get("project") !== "undefined" && isNaN(hashProjectId)) {
                    throw new Error("Project id in URL must be a number");
                }
                this.project_ids = [hashProjectId];

                return data.cache.get("projects_in_board", {
                    'id': this.board_id,
                    team_id: user.team_id[0]
                }).then((projectMap) => {
                    if (hashProjectId && projectMap.has(hashProjectId)) {
                        return;
                    } else if (hashProjectId) {
                        hash_service.delete("project");
                    }
                    //this.project_ids = user.team_ids[user.team_id[0]].project_ids;
                    const project_ids = [];
                    for (const project of projectMap) {
                        project_ids.push(project[0]);
                    }
                    this.project_ids = project_ids;
                });

            });
        },
        prepareShortcuts(list_id) {
            throw new Error("You should either implement this method or disable shortcuts by setting shouldRenderDragShortcuts to false");
        },
        renderDragShortcuts(shortcutPlace, shortcuts, sourceList) {
            this.shortcutContainer = $(`<div class="shortcut-container"></div>`);
            for (const shortcut of shortcuts) {
                if (shortcut.positioner) {
                    const positionerNode = $(`<div class="shortcut-positioner"/>`);
                    this.renderDragShortcutNode(shortcut, sourceList, TOP).appendTo(positionerNode);
                    const positionerNodeName = $(`<div class="shortcut-positioner-name">${shortcut.name}</div>`);
                    positionerNodeName.appendTo(positionerNode);
                    this.renderDragShortcutNode(shortcut, sourceList, BOTTOM).appendTo(positionerNode);
                    positionerNode.appendTo(this.shortcutContainer);
                } else {
                    const shortcutNode = this.renderDragShortcutNode(shortcut, sourceList);
                    shortcutNode.appendTo(this.shortcutContainer);
                }
            }
            shortcutPlace.append(this.shortcutContainer);
        },
        renderDragShortcutNode(shortcut, sourceList, position) {
            const name = position || shortcut.name;
            const positionData = position ? `data-shortcut-position="${position}"` : "";

            const shortcutNode = $(`<div class="shortcut-item agile-main-color lighten-5" ${positionData} data-shortcut-id="${shortcut.id}">${name}</div>`);
            shortcutNode[0].addEventListener("dragenter", (evt) => {
                $(evt.target).addClass("hover");
            }, false);
            shortcutNode[0].addEventListener("dragleave", (evt) => {
                $(evt.target).removeClass("hover");
            }, false);
            Sortable.create(shortcutNode[0], {
                group: "backlog",
                onAdd: function (evt) {
                    if (position) {
                        if (position === TOP) {
                            sourceList.moveToTop(evt);
                        } else if (position === BOTTOM) {
                            sourceList.moveToBottom(evt);
                        }
                    } else {
                        this.dragShortcutCallback(evt.item, sourceList, shortcut);
                    }
                }.bind(this)
            });
            return shortcutNode;
        },
        dragShortcutCallback(item, sourceList, shortcut) {
            const taskId = parseInt(item.dataset.id, 10);
            const itemWidget = this.taskWidgetItemMap.get(taskId);
            const newListWidget = sourceList._getNewListWidget(shortcut.id);
            // sourceList._setNewItemList(itemWidget, newListWidget);

            // insert to the bottom of the non-backlog list or in the beginning of backlog list
            const newPosition = shortcut.id ? newListWidget.list.size : 0;
            itemWidget.set_list(newListWidget, sourceList.getNewOrder(0, newPosition, shortcut.id, false));
        },

        _onProjectTaskWrite(id, vals, payload, record) {
            const editPromise = record && record._edit("check") ? record._edit() : $.when();
            editPromise.then(() => {
                this.removeTask(id, true);
                this.addTask(id, payload);
                if (this.rightSideWidget && this.rightSideWidget.id === id) {
                    // Since trigger_up wraps event arguments in data object, here I mimic that behaviour.
                    this.trigger("open_right_side", {
                        data: {
                            WidgetClass: TaskWidget,
                            options: {id, isQuickDetailView: true}
                        }
                    });
                }
            });
        },
        _onProjectTaskCreate(id, vals, payload) {
            this.addTask(id, payload);
        },
        _onProjectTaskUnlink(id, payload) {
            this.removeTask(id, true, payload);
            if (this.rightSideWidget && this.rightSideWidget.id === id) {
                this.rightSideWidget.destroy(true);
                delete this.rightSideWidget;
            }
        },
        // CUSTOM EVENT HENDLERS
        _onDragStart(evt) {
            if (this.shouldRenderDragShortcuts) {
                const shortcutPlace = this.$("#backlog-view");
                const shortcuts = this.prepareShortcuts(evt.target.$el.data("id"));
                this.renderDragShortcuts(shortcutPlace, shortcuts, evt.target);
            }
        },
        _onDragEnd(evt) {
            this.shortcutContainer && this.shortcutContainer.remove();
            if (!evt.data.sortableEvent.target || evt.data.sortableEvent.target.classList.contains("shortcut-item")) {
                evt.data.sortableEvent.preventDefault();
            }
        },
        _onBacklogListChanged(evt) {
            this.backlogNode.find(".task-count").text((evt.data.size || 0) + " of " + evt.data.total + " " + pluralize("issue", evt.data.total));
            Waypoint.refreshAll();
        },
        _onOpenLinkModal(evt) {
            if (!evt.data.id) {
                throw new Error("Event payload must contain id of task for wich link should be created");
            }
            task_service.getRecord(evt.data.id).then((record) => {
                if (!record) {
                    throw new Error("Task doens't exist");
                }
                const modal = new AgileModals.LinkItemModal(this, {
                    task: record,
                    task_ids: [...this.allTaskIds()],
                });
                modal.appendTo($("body"));
            });
        },
        _onOpenNewItemModal(evt) {
            const options = {
                projects: this.projects
            };
            Object.assign(options, evt.data);
            const newItemModal = new AgileModals.NewItemModal(this, options);
            newItemModal.appendTo($("body"));
        },
        _onLoadingMoreItemsStarted() {
            this.loadingMoreItems = true;
            this.$(".list-preloader").show();
        },
        _onLoadingMoreItemsFinished() {
            this.loadingMoreItems = false;
            if (this.backlogTaskList.shouldLoadMore() && this.$(".master-list").height() / this.$el.height() < 1.1) {
                this.backlogTaskList.loadMoreItems();
            } else {
                this.$(".list-preloader").hide();
            }
        },
        _onAddNewItemClick() {
            const defaults = {
                project: this.projects.find((p) => p.id === hash_service.get("project"))
            };
            this.trigger_up("open_new_item_modal", {
                currentProjectId: parseInt(hash_service.get("project"), 10) || null,
                focus: "name",
                defaults,
                beforeHook: this.setNewTaskOrder.bind(this),
            });
        }
        // CUSTOM EVENT HENDLERS END
    });

    return {
        AbstractBacklogView,
        BacklogList,
        SimpleBacklogTaskItem,
        NonBacklogList
    };
});
